"""Time series data for global or sink quantites.

This module contains a function for loading global quantities and sink
particle time series data typical of Phantom simulations. These files
track averaged quantities that are more frequently output than snapshot
files.
"""

from __future__ import annotations

import warnings
from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple, Union

from pandas import DataFrame

from .._logging import logger
from ._phantom_ev import load_data_from_file as load_data_from_file_phantom
from ._phantom_ev import time_series_units as time_series_units_phantom

if TYPE_CHECKING:
    from .simulation import Simulation


def load_ev(
    filenames: Union[str, Path, Tuple[str], Tuple[Path], List[str], List[Path]],
    data_source: str = 'Phantom',
    config: Union[str, Path] = None,
) -> DataFrame:
    """Load time series data from file(s).

    Time series files track global quantities, such as energy, momentum,
    and density, over time. The time increments in these files is
    smaller than the snapshot file output time. These files are
    typically stored as text files.

    The data is stored as a pandas DataFrame.

    Parameters
    ----------
    filename(s)
        Collection of paths to time series file(s) in chronological
        order. These should all contain the same columns.
    data_source : optional
        The code used to produce the data. Default is 'Phantom'.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    dataframe
        A pandas DataFrame with the time series data.

    Examples
    --------
    Reading a single time series file into a pandas DataFrame.

    >>> file_name = 'simulation.ev'
    >>> ts = plonk.load_time_series(file_name)

    Reading a collection of time series files into a pandas DataFrame.

    >>> file_names = ('sim01.ev', 'sim02.ev', 'sim03.ev')
    >>> ts = plonk.load_time_series(file_names)
    """
    msg = (
        'load_ev is deprecated and will be removed in v0.7.4, '
        'please use load_time_series instead'
    )
    logger.warning(msg)
    warnings.warn(msg, DeprecationWarning)
    return load_time_series(filenames=filenames, data_source=data_source, config=config)


def load_time_series(
    filenames: Union[str, Path, Tuple[str], Tuple[Path], List[str], List[Path]],
    data_source: str = 'Phantom',
    config: Union[str, Path] = None,
) -> DataFrame:
    """Load time series data from file(s).

    Time series files track global quantities, such as energy, momentum,
    and density, over time. The time increments in these files is
    smaller than the snapshot file output time. These files are
    typically stored as text files.

    The data is stored as a pandas DataFrame.

    Parameters
    ----------
    filename(s)
        Collection of paths to time series file(s) in chronological
        order. These should all contain the same columns.
    data_source : optional
        The code used to produce the data. Default is 'Phantom'.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    dataframe
        A pandas DataFrame with the time series data.

    Examples
    --------
    Reading a single time series file into a pandas DataFrame.

    >>> file_name = 'simulation.ev'
    >>> ts = plonk.load_time_series(file_name)

    Reading a collection of time series files into a pandas DataFrame.

    >>> file_names = ('sim01.ev', 'sim02.ev', 'sim03.ev')
    >>> ts = plonk.load_time_series(file_names)
    """
    if data_source == 'Phantom':
        return load_data_from_file_phantom(filenames=filenames, config=config)
    raise ValueError('Cannot determine code used to produce time series data')


def time_series_units(
    sim: Simulation, data_source: str = 'Phantom', config: Union[str, Path] = None
):
    """Get units of time series data from Simulation object.

    Parameters
    ----------
    sim
        The Simulation object.
    data_source : optional
        The code used to produce the data. Default is 'Phantom'.
    config : optional
        The path to a Plonk config.toml file.

    Returns
    -------
    Dict
    """
    if data_source == 'Phantom':
        return time_series_units_phantom(sim=sim, config=config)
    raise ValueError('Cannot determine code used to produce time series data')



"""Snapshot data access and manipulation.

It contains Plonk the implementation of smoothed particle hydrodynamics
snapshot file.
"""

from pathlib import Path
from typing import Union

from ..utils.strings import is_documented_by
from .snap import Sinks, Snap, SnapLike, SubSnap


@is_documented_by(Snap.load_snap)
def load_snap(
    filename: Union[str, Path],
    data_source: str = 'phantom',
    config: Union[str, Path] = None,
):
    return Snap().load_snap(filename=filename, data_source=data_source, config=config)


__all__ = [
    'Sinks',
    'Snap',
    'SnapLike',
    'SubSnap',
    'load_snap',
]



"""Utils for math."""

import numpy as np
from numpy import ndarray

from .._units import Quantity


def cross(x, y, **kwargs):
    """Cross product.

    Parameters
    ----------
    x, y
        The two arrays (N, 3) to take the cross product of. Can be
        ndarray or pint Quantity.
    **kwargs
        Keyword arguments to pass to np.cross.

    Returns
    -------
    ndarray
        The cross product of x and y.
    """
    if isinstance(x, Quantity):
        return np.cross(x.magnitude, y.magnitude, **kwargs) * x.units * y.units
    return np.cross(x, y, **kwargs)


def norm(x, **kwargs):
    """Norm of a vector.

    Parameters
    ----------
    x
        The arrays (N, 3) to take the norm of. Can be ndarray or pint
        Quantity.
    **kwargs
        Keyword arguments to pass to np.linalg.norm.

    Returns
    -------
    ndarray
        The norm of x.
    """
    if isinstance(x, Quantity):
        return np.linalg.norm(x.magnitude, **kwargs) * x.units
    return np.linalg.norm(x, **kwargs)


def average(x, weights, **kwargs):
    """Average.

    Parameters
    ----------
    x
        The array (N,) to take the norm of. Can be ndarray or pint
        Quantity.
    weights
        The weights for averaging.
    **kwargs
        Keyword arguments to pass to np.average.

    Returns
    -------
    ndarray
        The average of x.
    """
    if isinstance(x, Quantity):
        return np.average(x.magnitude, weights=weights.magnitude, **kwargs) * x.units
    return np.average(x, weights=weights, **kwargs)


def distance_from_plane(
    x: ndarray, y: ndarray, z: ndarray, normal: ndarray, height: float = 0
) -> ndarray:
    """Calculate distance from a plane.

    Parameters
    ----------
    x
        The x-positions.
    y
        The y-positions.
    z
        The z-positions.
    normal
        The normal vector describing the plane (x, y, z).
    height
        The height of the plane above the origin.

    Return
    ------
    The distance from the plane of each point.
    """
    a, b, c = normal
    d = height
    return np.abs((a * x + b * y + c * z + d) / np.sqrt(a ** 2 + b ** 2 + c ** 2))



"""SPH simulation data.

It contains Plonk the implementation of smoothed particle hydrodynamics
combined simulation data. This includes snapshots and time series data.
"""

from .simulation import Simulation, load_simulation
from .time_series import load_time_series

__all__ = ['Simulation', 'load_time_series', 'load_simulation']



"""Animations of visualizations."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Union

import matplotlib.pyplot as plt
from matplotlib import animation as _animation

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

from .._logging import logger
from .._units import units as plonk_units
from . import visualization as viz

if TYPE_CHECKING:
    from ..analysis.profile import Profile
    from ..snap.snap import SnapLike


def animate(
    filename: Union[str, Path],
    *,
    snaps: List[SnapLike] = None,
    profiles: List[Profile] = None,
    quantity: str = None,
    x: str = None,
    y: str = None,
    fig: Any = None,
    adaptive_colorbar: bool = True,
    adaptive_limits: bool = True,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Animate a Snap or Profile visualization.

    Pass in a list of Snap objects or Profile objects. If snaps are
    passed in and the quantity is None, then the animation will be of
    particle plots, otherwise it will be of images. If profiles are
    passed in the animation will be of profiles.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    snaps
        A list of Snap objects to animate.
    profiles
        A list of Profile objects to animate.
    quantity : optional
        The quantity to visualize. Must be a string to pass to Snap.
    x : optional
        The quantity for the x-axis. Must be a string to pass to Snap.
        For interpolated plots must be 'x', 'y', or 'z'.
    y : optional
        The quantity for the y-axis. Must be a string to pass to Snap.
        For interpolated plots must be 'x', 'y', or 'z'.
    fig : optional
        A matplotlib Figure object to animate. If None, generate a new
        Figure.
    adaptive_colorbar : optional
        If True, adapt colorbar range during animation. If False, the
        colorbar range is fixed by the initial image. Default is True.
    adaptive_limits : optional
        If True, adapt plot limits during animation. If False, the plot
        limits are fixed by the initial plot. Default is True.
    text : optional
        List of strings to display per snap.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to visualize.image.

    Returns
    -------
    anim
        The matplotlib FuncAnimation object.

    Examples
    --------
    Make an image animation of projected density.

    >>> units = {
    ...     'position': 'au',
    ...     'density': 'g/cm^3',
    ...     'projection': 'cm'
    ... }
    >>> plonk.animate(
    ...     filename='animation.mp4',
    ...     snaps=snaps,
    ...     quantity='density',
    ...     units=units,
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )

    Make a particle animation of x vs density.

    >>> units = {'position': 'au', 'density': 'g/cm^3'}
    >>> plonk.animate(
    ...     filename='animation.mp4',
    ...     snaps=snaps,
    ...     x='x',
    ...     y='density',
    ...     units=units,
    ...     adaptive_limits=False,
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )

    Make a profile animation of radius vs surface density.

    >>> units={'position': 'au', 'surface_density': 'g/cm^2'}
    >>> plonk.animate(
    ...     filename='animation.mp4',
    ...     profiles=profiles,
    ...     x='radius',
    ...     y='surface_density',
    ...     units=units,
    ...     adaptive_limits=False,
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )
    """
    if snaps is not None:
        if profiles is not None:
            raise ValueError('Cannot pass in snaps and profiles')
        if quantity is None:
            anim = animation_particles(
                filename=filename,
                snaps=snaps,
                x=x,
                y=y,
                fig=fig,
                adaptive_limits=adaptive_limits,
                text=text,
                text_kwargs=text_kwargs,
                func_animation_kwargs=func_animation_kwargs,
                save_kwargs=save_kwargs,
                **kwargs,
            )
        else:
            anim = animation_images(
                filename=filename,
                snaps=snaps,
                quantity=quantity,
                fig=fig,
                adaptive_colorbar=adaptive_colorbar,
                text=text,
                text_kwargs=text_kwargs,
                func_animation_kwargs=func_animation_kwargs,
                save_kwargs=save_kwargs,
                **kwargs,
            )
    elif profiles is not None:
        if x is None:
            raise ValueError('x must be passed in for profile animations')
        if y is None:
            raise ValueError('y must be passed in for profile animations')
        anim = animation_profiles(
            filename=filename,
            profiles=profiles,
            x=x,
            y=y,
            fig=fig,
            adaptive_limits=adaptive_limits,
            text=text,
            text_kwargs=text_kwargs,
            func_animation_kwargs=func_animation_kwargs,
            save_kwargs=save_kwargs,
            **kwargs,
        )
    else:
        raise ValueError('Must pass in snaps or profiles')

    return anim


def animation_images(
    *,
    filename: Union[str, Path],
    snaps: List[SnapLike],
    quantity: str,
    fig: Any = None,
    adaptive_colorbar: bool = True,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation of images.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    snaps
        A list of Snap objects to animate.
    quantity
        The quantity to visualize. Must be a string to pass to Snap.
    fig : optional
        A matplotlib Figure object to animate. If None, generate a new
        Figure.
    adaptive_colorbar : optional
        If True, adapt colorbar range during animation. If False, the
        colorbar range is fixed by the initial image. Default is True.
    text : optional
        List of strings to display per snap.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to visualize.image.

    Returns
    -------
    anim
        The matplotlib FuncAnimation object.

    Examples
    --------
    Make an animation of projected density.

    >>> animation_images(
    ...     filename='animation.mp4',
    ...     snaps=snaps,
    ...     quantity='density',
    ...     units={'position': 'au', 'density': 'g/cm^3'},
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )
    """
    filepath = Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    if fig is None:
        ax = snaps[0].image(quantity=quantity, **kwargs)
        fig = ax.figure
        image = ax.images[0]
    else:
        ax = fig.axes[0]
        image = ax.images[0]

    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    if tqdm is not None:
        pbar = tqdm(total=len(snaps))
    else:
        logger.info(
            'progress bar not available\ntry pip install tqdm --or-- conda install tqdm'
        )

    x = kwargs.get('x', 'x')
    y = kwargs.get('y', 'y')
    interp = kwargs.get('interp', 'projection')
    slice_normal = kwargs.get('slice_normal')
    slice_offset = kwargs.get('slice_offset')
    extent = kwargs.get('extent')
    units = kwargs.get('units')
    weighted = kwargs.get('weighted')
    num_pixels = kwargs.get('num_pixels')

    def animate(idx):
        if tqdm is not None:
            pbar.update(n=1)
        snap = snaps[idx]
        interp_data, _extent, _units = viz._interpolated_data(
            snap=snap,
            quantity=quantity,
            x=x,
            y=y,
            interp=interp,
            slice_normal=slice_normal,
            slice_offset=slice_offset,
            extent=extent,
            units=units,
            weighted=weighted,
            num_pixels=num_pixels,
        )
        image.set_data(interp_data)
        image.set_extent(_extent)
        image.axes.set_xlim(_extent[:2])
        image.axes.set_ylim(_extent[2:])
        if adaptive_colorbar:
            vmin = kwargs.get('vmin', interp_data.min())
            vmax = kwargs.get('vmax', interp_data.max())
            image.set_clim(vmin, vmax)
        if text is not None:
            _text.set_text(text[idx])
        return [image]

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(snaps), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()
    if tqdm is not None:
        pbar.close()

    return anim


def animation_profiles(
    *,
    filename: Union[str, Path],
    profiles: List[Profile],
    x: str,
    y: Union[str, List[str]],
    fig: Any = None,
    adaptive_limits: bool = True,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation of a profile.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    profiles
        A list of Profile objects to animate.
    x
        The quantity for the x-axis. Must be a string to pass to Snap.
    y
        The quantity for the y-axis, or list of quantities. Must be a
        string (or list of strings) to pass to Snap.
    fig : optional
        A matplotlib Figure object to animate. If None, generate a new
        Figure.
    adaptive_limits : optional
        If True, adapt plot limits during animation. If False, the plot
        limits are fixed by the initial plot. Default is True.
    text : optional
        List of strings to display per profile plot.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to Profile.plot function.

    Returns
    -------
    anim
        The matplotlib FuncAnimation object.

    Examples
    --------
    Make an animation of radius vs surface density.

    >>> animation_profiles(
    ...     filename='animation.mp4',
    ...     profiles=profiles,
    ...     x='radius',
    ...     y='surface_density',
    ...     units={'position': 'au', 'surface_density': 'g/cm^2'},
    ...     adaptive_limits=False,
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )
    """
    filepath = Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    if isinstance(y, str):
        ys = [y]
    elif isinstance(y, (tuple, list)):
        ys = y
    else:
        raise ValueError('y must be a string or list of strings')

    prof = profiles[0]
    if fig is None:
        ax = prof.plot(x=x, y=ys, **kwargs)
        fig = ax.figure
        lines = ax.lines
    else:
        ax = fig.axes[0]
        lines = [line for line in ax.lines]

    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    if tqdm is not None:
        pbar = tqdm(total=len(profiles))
    else:
        logger.info(
            'progress bar not available\ntry pip install tqdm --or-- conda install tqdm'
        )

    units = kwargs.get('units')
    if units is not None:
        xunit = plonk_units(units.get(prof.base_profile_name(x), str(prof[x].units)))
        yunits = [
            plonk_units(units.get(prof.base_profile_name(y), str(prof[y].units)))
            for y in ys
        ]
    else:
        xunit = prof[x].units
        yunits = [prof[y].units for y in ys]

    def animate(idx):
        if tqdm is not None:
            pbar.update(n=1)
        profile = profiles[idx]
        xlim, ylim = (0, 0), (0, 0)
        _x = (profile[x] / xunit).to_base_units().magnitude
        for line, y, yunit in zip(lines, ys, yunits):
            _y = (profile[y] / yunit).to_base_units().magnitude
            line.set_data(_x, _y)
            if adaptive_limits:
                ax = line.axes
                xlim, ylim = _get_range(_x, xlim), _get_range(_y, ylim)
                ax.set(xlim=xlim, ylim=ylim)
        if text is not None:
            _text.set_text(text[idx])
        return lines

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(profiles), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()
    if tqdm is not None:
        pbar.close()

    return anim


def animation_particles(
    *,
    filename: Union[str, Path],
    snaps: List[SnapLike],
    fig: Any = None,
    adaptive_limits: bool = True,
    text: List[str] = None,
    text_kwargs: Dict[str, Any] = {},
    func_animation_kwargs: Dict[str, Any] = {},
    save_kwargs: Dict[str, Any] = {},
    **kwargs,
):
    """Generate an animation of particle plots.

    Parameters
    ----------
    filename
        The file name to save the animation to.
    snaps
        A list of Snap objects to animate.
    fig : optional
        A matplotlib Figure object to animate. If None, generate a new
        Figure.
    adaptive_limits : optional
        If True, adapt plot limits during animation. If False, the plot
        limits are fixed by the initial plot. Default is True.
    text : optional
        List of strings to display per snap.
    text_kwargs : optional
        Keyword arguments to pass to matplotlib text.
    func_animation_kwargs : optional
        Keyword arguments to pass to matplotlib FuncAnimation.
    save_kwargs : optional
        Keyword arguments to pass to matplotlib Animation.save.
    **kwargs
        Arguments to pass to visualize.plot.

    Returns
    -------
    anim
        The matplotlib FuncAnimation object.

    Examples
    --------
    Make an animation of x vs density on the particles.

    >>> animation_particles(
    ...     filename='animation.mp4',
    ...     snaps=snaps,
    ...     x='x',
    ...     y='density',
    ...     adaptive_limits=False,
    ...     save_kwargs={'fps': 10, 'dpi': 300},
    ... )
    """
    filepath = Path(filename)
    if filepath.suffix != '.mp4':
        raise ValueError('filename should end in ".mp4"')

    if fig is None:
        ax = snaps[0].plot(**kwargs)
        fig = ax.figure
        lines = ax.lines
    else:
        ax = fig.axes[0]
        lines = [line for line in ax.lines]

    if text is not None:
        _text = ax.text(
            0.9, 0.9, text[0], ha='right', transform=ax.transAxes, **text_kwargs
        )

    if tqdm is not None:
        pbar = tqdm(total=len(snaps))
    else:
        logger.info(
            'progress bar not available\ntry pip install tqdm --or-- conda install tqdm'
        )

    x = kwargs.get('x', 'x')
    y = kwargs.get('y', 'y')
    c = kwargs.get('c')
    s = kwargs.get('s')
    units = kwargs.get('units')

    def animate(idx):
        snap = snaps[idx]
        if tqdm is not None:
            pbar.update(n=1)
        xlim, ylim = (0, 0), (0, 0)

        if c is None and s is None:
            try:
                subsnaps = snap.subsnaps_as_list(squeeze=False)
            except AttributeError:
                subsnaps = [snap]
        else:
            subsnaps = [snap]

        for line, subsnap in zip(lines, subsnaps):
            _x, _y, _c, _s, _units = viz._plot_data(
                snap=subsnap, x=x, y=y, c=c, s=s, units=units
            )
            line.set_data(_x, _y)
            if adaptive_limits:
                xlim, ylim = _get_range(_x, xlim), _get_range(_y, ylim)
                line.axes.set(xlim=xlim, ylim=ylim)
        if text is not None:
            _text.set_text(text[idx])
        return lines

    anim = _animation.FuncAnimation(
        fig, animate, frames=len(snaps), **func_animation_kwargs
    )
    anim.save(filepath, extra_args=['-vcodec', 'libx264'], **save_kwargs)
    plt.close()
    if tqdm is not None:
        pbar.close()

    return anim


def _get_range(vals, vlim):
    vmin, vmax = vals.min(), vals.max()
    dv = 0.05 * (vmax - vmin)
    return (
        vmin - dv if vmin - dv < vlim[0] else vlim[0],
        vmax + dv if vmax + dv > vlim[1] else vlim[1],
    )



"""Analysis for accretion discs."""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Union

import numpy as np
from numpy import ndarray

from .._units import Quantity
from .._units import units as plonk_units
from ..utils.math import cross, norm
from .total import angular_momentum, center_of_mass

if TYPE_CHECKING:
    from ..snap.snap import SnapLike

G = (1 * plonk_units.newtonian_constant_of_gravitation).to_base_units()

# Derived quantities require some arrays already present
# Ignoring type, sub_type, (optional) smoothing_length
array_requires = {
    'eccentricity': ['position', 'velocity'],
    'inclination': ['position', 'velocity'],
    'keplerian_frequency': ['position'],
    'semi_major_axis': ['position', 'velocity'],
    'stokes_number': ['position', 'stopping_time'],
}

# Arrays which represent quantities with x-, y-, z-components in space
vector_arrays: List[str] = []

# Arrays which represent dust quantities with columns per dust species
dust_arrays = ['stokes_number']


def eccentricity(
    snap: SnapLike,
    central_body: Dict[str, Quantity] = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the eccentricity.

    Parameters
    ----------
    snap
        The Snap object.
    central_body : optional
        A dictionary with the mass, position, and velocity (as Pint
        quantities) of the central body around which the particles are
        orbiting. If None, attempt to read from
        snap.properties['central_body'].
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The eccentricity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    if central_body is None:
        try:
            central_body = snap.properties['central_body']
        except KeyError:
            raise ValueError(
                'must pass in central_body or '
                'set both on snap with snap.set_central_body'
            )

    mu = (G * central_body['mass']).to_reduced_units()
    r = pos - central_body['position']
    v = vel - central_body['velocity']

    h = norm(cross(r, v), axis=1)
    eps_k = 1 / 2 * norm(v, axis=1) ** 2
    eps_p = -mu / norm(r, axis=1)
    eps = eps_k + eps_p

    return np.sqrt(1 + 2 * eps * (h / mu) ** 2)


def inclination(
    snap: SnapLike,
    central_body: Dict[str, Quantity] = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the inclination.

    Parameters
    ----------
    snap
        The Snap object.
    central_body : optional
        A dictionary with the mass, position, and velocity (as Pint
        quantities) of the central body around which the particles are
        orbiting. If None, attempt to read from
        snap.properties['central_body'].
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The inclination on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    if central_body is None:
        try:
            central_body = snap.properties['central_body']
        except KeyError:
            raise ValueError(
                'must pass in central_body or '
                'set both on snap with snap.set_central_body'
            )

    r = pos - central_body['position']
    v = vel - central_body['velocity']
    h = cross(r, v)

    return np.arccos(h[:, 2] / norm(h, axis=1))


def keplerian_frequency(
    snap: SnapLike,
    central_body: Dict[str, Quantity] = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the Keplerian orbital frequency.

    Parameters
    ----------
    snap
        The Snap object.
    central_body : optional
        A dictionary with the mass, position, and velocity (as Pint
        quantities) of the central body around which the particles are
        orbiting. If None, attempt to read from
        snap.properties['central_body'].
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The Keplerian frequency on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    if central_body is None:
        try:
            central_body = snap.properties['central_body']
        except KeyError:
            raise ValueError(
                'must pass in central_body or '
                'set both on snap with snap.set_central_body'
            )

    mu = (G * central_body['mass']).to_reduced_units()
    r = norm(pos - central_body['position'], axis=1)

    return np.sqrt(mu / r ** 3)


def semi_major_axis(
    snap: SnapLike,
    central_body: Dict[str, Quantity] = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the semi-major axis.

    Parameters
    ----------
    snap
        The Snap object.
    central_body : optional
        A dictionary with the mass, position, and velocity (as Pint
        quantities) of the central body around which the particles are
        orbiting. If None, attempt to read from
        snap.properties['central_body'].
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The semi-major axis on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    if central_body is None:
        try:
            central_body = snap.properties['central_body']
        except KeyError:
            raise ValueError(
                'must pass in central_body or '
                'set both on snap with snap.set_central_body'
            )

    mu = (G * central_body['mass']).to_reduced_units()
    r = norm(pos - central_body['position'], axis=1)
    v = vel - central_body['velocity']

    eps_k = 1 / 2 * norm(v, axis=1) ** 2
    eps_p = -mu / r
    eps = eps_k + eps_p

    return -mu / (2 * eps)


def stokes_number(
    snap: SnapLike,
    central_body: Dict[str, Quantity] = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the Stokes number.

    Parameters
    ----------
    snap
        The Snap object.
    central_body : optional
        A dictionary with the mass, position, and velocity (as Pint
        quantities) of the central body around which the particles are
        orbiting. If None, attempt to read from
        snap.properties['central_body'].
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The Stokes number on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        t_s: Quantity = snap['stopping_time'][h > 0]
    else:
        pos = snap['position']
        t_s = snap['stopping_time']

    if central_body is None:
        try:
            central_body = snap.properties['central_body']
        except KeyError:
            raise ValueError(
                'must pass in central_body or '
                'set both on snap with snap.set_central_body'
            )

    mu = (G * central_body['mass']).to_reduced_units()
    r = norm(pos - central_body['position'], axis=1)

    Omega_k = np.sqrt(mu / r ** 3)

    return t_s * Omega_k[:, np.newaxis]


def unit_normal(snap: SnapLike, sinks: Union[bool, List[int]] = False) -> ndarray:
    """Calculate unit normal to plane of rotation.

    I.e. calculate a unit angular momentum vector.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    ndarray
        A unit angular momentum vector.
    """
    origin = center_of_mass(snap=snap, sinks=sinks)
    L = angular_momentum(snap=snap, sinks=sinks, origin=origin).magnitude
    return L / np.linalg.norm(L)


def rotate_face_on(snap: SnapLike, sinks: Union[bool, List[int]] = False) -> SnapLike:
    """Rotate to face-on with the angular momentum vector.

    I.e. rotate such that the angular momentum points in the
    z-direction.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    Snap
        The rotated Snap.
    """
    vec = unit_normal(snap=snap, sinks=sinks)
    z_axis = np.array([0, 0, 1])
    axis = cross(vec, z_axis)
    angle = np.arccos(np.dot(vec, z_axis))
    return snap.rotate(axis=axis, angle=angle)


def rotate_edge_on(snap: SnapLike, sinks: Union[bool, List[int]] = False) -> SnapLike:
    """Rotate to edge-on with the angular momentum vector.

    I.e. rotate such that the angular momentum points in the
    y-direction.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    Snap
        The rotated Snap.
    """
    vec = unit_normal(snap=snap, sinks=sinks)
    y_axis = np.array([0, 1, 0])
    axis = cross(vec, y_axis)
    angle = np.arccos(np.dot(vec, y_axis))
    return snap.rotate(axis=axis, angle=angle)


def position_angle(snap: SnapLike) -> Quantity:
    """Calculate the disc position angle.

    The position angle is taken from the x-axis in the xy-plane. It
    defines a unit vector around which the snap is inclined.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    Quantity
        The disc position angle.
    """
    angmom = angular_momentum(snap=snap, sinks=False)
    if isinstance(angmom, Quantity):
        pi_2 = np.pi / 2 * Quantity('radian')
    else:
        pi_2 = np.pi / 2
    return np.arctan2(angmom[1], angmom[0]) + pi_2


def inclination_angle(snap: SnapLike) -> Quantity:
    """Calculate the disc inclination.

    The inclination is calculated by taking the angle between the
    angular momentum vector and the z-axis, with the angular momentum
    calculated with respect to the center of mass.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    Quantity
        The disc inclination.
    """
    angmom = angular_momentum(snap=snap, sinks=False)
    return np.arccos(angmom[2] / norm(angmom))



"""Utils for snaps."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, List, Union

from .._config import read_config
from .._units import units as plonk_units

if TYPE_CHECKING:
    from ..snap.snap import SnapLike


def gravitational_constant_in_code_units(snap: SnapLike) -> float:
    """Gravitational constant in code units.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    float
        The gravitational constant in code units.
    """
    G = plonk_units.newtonian_constant_of_gravitation
    G_units = (
        snap.code_units['length'] ** 3
        / snap.code_units['mass']
        / snap.code_units['time'] ** 2
    )
    G = (G / G_units).to_base_units().magnitude
    return G


def add_aliases(snap: SnapLike, filename: Union[str, Path] = None):
    """Add array aliases to a Snap.

    Parameters
    ----------
    snap
        The Snap object.
    config : optional
        The path to a Plonk config.toml file. If None, use the default
        file.
    """
    conf = read_config(filename=filename)
    for key, val in conf['arrays']['aliases'].items():
        snap.add_alias(key, val)


def dust_array_names(
    name: str, num_dust_species: int, add_gas: bool = False
) -> List[str]:
    """List dust array names.

    Parameters
    ----------
    name
        The base array name, e.g. "dust_density" or "stopping_time".
    num_dust_species
        The number of dust species.
    add_gas
        If True add the gas version of the dust name.

    Returns
    -------
    List
        A list of array names with appropriate suffixes.

    Examples
    --------
    Get the dust density strings.

    >>> dust_name_list('dust_density', 5)
    ['dust_density_001',
     'dust_density_002',
     'dust_density_003',
     'dust_density_004',
     'dust_density_005']

    Get the dust density strings with gas.

    >>> dust_name_list(name='dust_density', num_dust_species=5, add_gas=True)
    ['gas_density',
     'dust_density_001',
     'dust_density_002',
     'dust_density_003',
     'dust_density_004',
     'dust_density_005']
    """
    names = list()
    if add_gas:
        names.append(f'{name.replace("dust", "gas")}')
    names += [f'{name}_{n+1:03}' for n in range(num_dust_species)]
    return names


def vector_array_names(name: str, add_mag: bool = False) -> List[str]:
    """List vector array names.

    Parameters
    ----------
    name
        The base array name, e.g. "angular_momentum".
    add_mag
        If True add the magnitude of the array.

    Returns
    -------
    List
        A list of array names with appropriate suffixes.

    Examples
    --------
    Get the angular momentum strings.

    >>> vector_name_list('angular_momentum')
    ['angular_momentum_x',
     'angular_momentum_y',
     'angular_momentum_z']

    Get the angular momentum strings with magnitude.

    >>> vector_name_list(name='angular_momentum', add_mag=True)
    ['angular_momentum_x',
     'angular_momentum_y',
     'angular_momentum_z',
     'angular_momentum_mag']
    """
    names = [f'{name}_{x}' for x in ['x', 'y', 'z']]
    if add_mag:
        names.append(f'{name}_mag')
    return names



"""Calculate extra quantities on the particles."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import numpy as np

from .._logging import logger
from .._units import Quantity
from .._units import units as plonk_units
from ..utils.math import cross, norm

if TYPE_CHECKING:
    from ..snap.snap import Sinks, SnapLike

ORIGIN = (0, 0, 0) * plonk_units('meter')
MOLECULAR_HYDROGEN_WEIGHT = 2.381

# Derived quantities require some arrays already present
# Ignoring type, sub_type, (optional) smoothing_length
array_requires = {
    'angular_momentum': ['mass', 'position', 'velocity'],
    'angular_velocity': ['position', 'velocity'],
    'azimuthal_angle': ['position'],
    'dust_density': ['density', 'dust_fraction'],
    'dust_mass': ['dust_fraction', 'mass'],
    'gas_density': ['density', 'dust_fraction'],
    'gas_fraction': ['dust_fraction'],
    'gas_mass': ['dust_fraction', 'mass'],
    'kinetic_energy': ['mass', 'velocity'],
    'momentum': ['mass', 'velocity'],
    'polar_angle': ['position'],
    'radius_cylindrical': ['position'],
    'radius_spherical': ['position'],
    'specific_angular_momentum': ['position', 'velocity'],
    'specific_kinetic_energy': ['velocity'],
    'temperature': ['sound_speed'],
    'velocity_radial_cylindrical': ['position', 'velocity'],
    'velocity_radial_spherical': ['position', 'velocity'],
}

# Arrays which represent quantities with x-, y-, z-components in space
vector_arrays = ['angular_momentum', 'momentum', 'specific_angular_momentum']

# Arrays which represent dust quantities with columns per dust species
dust_arrays = ['dust_density', 'dust_mass', 'stokes_number']


def angular_momentum(
    snap: Union[SnapLike, Sinks], origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular momentum as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The angular momentum on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        mass: Quantity = snap['mass'][h > 0]
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        pos = snap['position']
        vel = snap['velocity']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    if pos.ndim == 1:
        return mass * cross(pos, vel)
    return mass[:, np.newaxis] * cross(pos, vel)


def angular_velocity(
    snap: Union[SnapLike, Sinks], origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the angular velocity.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the angular velocity as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The angular velocity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    x, y = pos[:, 0], pos[:, 1]
    vx, vy = vel[:, 0], vel[:, 1]

    vphi = (x * vy - y * vx) / (x ** 2 + y ** 2)

    return vphi


def azimuthal_angle(
    snap: Union[SnapLike, Sinks], origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the azimuthal angle.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the azimuthal angle as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The azimuthal angle on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    x, y = pos[:, 0], pos[:, 1]
    return np.arctan2(y, x)


def dust_density(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the dust density per species.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The dust density per species on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        raise ValueError(
            'Dust method is "dust as separate sets of particles"\n'
            'To get dust_density create a SubSnap with snap.family("dust")\n'
            'and then access "density"'
        )

    else:
        raise ValueError('No dust available')

    density: Quantity = snap['density']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _dust_mass = density[:, np.newaxis] * _dust_fraction
        return _dust_mass[h > 0]
    return density[:, np.newaxis] * _dust_fraction


def dust_mass(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the dust mass per species.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The dust mass per species on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        raise ValueError(
            'Dust method is "dust as separate sets of particles"\n'
            'To get dust_mass create a SubSnap with snap.family("dust")\n'
            'and then access "mass"'
        )

    else:
        raise ValueError('No dust available')

    mass: Quantity = snap['mass']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        _dust_mass = mass[:, np.newaxis] * _dust_fraction
        return _dust_mass[h > 0]
    return mass[:, np.newaxis] * _dust_fraction


def gas_density(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the gas density.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas density on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']
        _gas_fraction = 1 - _dust_fraction.sum(axis=1)

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        raise ValueError(
            'Dust method is "dust as separate sets of particles"\n'
            'To get gas_density create a SubSnap with snap.family("gas")\n'
            'and then access "density"'
        )

    density: Quantity = snap['density']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        return (density * _gas_fraction)[h > 0]
    return density * _gas_fraction


def gas_fraction(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the gas fraction.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas fraction on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']
        _gas_fraction = 1 - _dust_fraction.sum(axis=1)

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        raise ValueError(
            'Dust method is "dust as separate sets of particles"\n'
            'The gas_fraction is 1 on gas particles, and 0 otherwise'
        )

    else:
        _gas_fraction = np.ones(len(snap))

    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        return _gas_fraction[h > 0]
    return _gas_fraction


def gas_mass(snap: SnapLike, ignore_accreted: bool = False) -> Quantity:
    """Calculate the gas mass.

    For dust/gas mixtures this is from the dust fraction.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas mass on the particles.
    """
    if snap.properties['dust_method'] == 'dust/gas mixture':
        _dust_fraction: Quantity = snap['dust_fraction']
        _gas_fraction = 1 - _dust_fraction.sum(axis=1)

    elif snap.properties['dust_method'] == 'dust as separate sets of particles':
        raise ValueError(
            'Dust method is "dust as separate sets of particles"\n'
            'To get gas_density create a SubSnap with snap.family("gas")\n'
            'and then access "mass"'
        )

    mass: Quantity = snap['mass']
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        return (mass * _gas_fraction)[h > 0]
    return mass * _gas_fraction


def kinetic_energy(
    snap: Union[SnapLike, Sinks], ignore_accreted: bool = False
) -> Quantity:
    """Calculate the kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The kinetic energy on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        mass: Quantity = snap['mass'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        vel = snap['velocity']

    return 1 / 2 * mass * norm(vel, axis=1) ** 2


def momentum(snap: Union[SnapLike, Sinks], ignore_accreted: bool = False) -> Quantity:
    """Calculate the momentum.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The linear momentum on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        mass: Quantity = snap['mass'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        mass = snap['mass']
        vel = snap['velocity']

    if vel.ndim == 1:
        return mass * vel
    return mass[:, np.newaxis] * vel


def polar_angle(
    snap: Union[SnapLike, Sinks], origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the polar angle.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the polar angle as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The azimuthal angle on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    r = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    return np.arccos(z / r)


def radius_cylindrical(
    snap: Union[SnapLike, Sinks],
    origin: Quantity = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the cylindrical radial distance.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the cylindrical radius as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial distance on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    x, y = pos[:, 0], pos[:, 1]

    return np.sqrt(x ** 2 + y ** 2)


def radius_spherical(
    snap: Union[SnapLike, Sinks],
    origin: Quantity = None,
    ignore_accreted: bool = False,
) -> Quantity:
    """Calculate the spherical radial distance.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the spherical radius as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial distance on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
    else:
        pos = snap['position']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]

    return np.sqrt(x ** 2 + y ** 2 + z ** 2)


def specific_angular_momentum(
    snap: Union[SnapLike, Sinks], origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the specific angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the specific angular momentum
        as a Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The specific angular momentum on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    return cross(pos, vel)


def specific_kinetic_energy(
    snap: Union[SnapLike, Sinks], ignore_accreted: bool = False
) -> Quantity:
    """Calculate the specific kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The specific kinetic energy on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        vel: Quantity = snap['velocity'][h > 0]
    else:
        vel = snap['velocity']

    return 1 / 2 * norm(vel, axis=1) ** 2


def temperature(
    snap: SnapLike, molecular_weight: float = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the gas temperature.

    Parameters
    ----------
    snap
        The Snap object.
    molecular_weight
        The gas molecular weight in gram / mole. E.g. 2.381 for
        molecular hydrogen.
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The gas temperature on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        cs: Quantity = snap['sound_speed'][h > 0]
    else:
        cs = snap['sound_speed']

    gamma = snap.properties['adiabatic_index']

    molecular_weight = snap.properties.get('molecular_weight')
    if molecular_weight is None:
        logger.warning(
            'molecular_weight not set, assuming 2.381 for molecular hydrogen\n'
            'use set_molecular_weight to set on Snap'
        )
        molecular_weight = MOLECULAR_HYDROGEN_WEIGHT

    molecular_weight = molecular_weight * plonk_units('gram / mole')
    specific_gas_constant = (plonk_units.R / molecular_weight).to_base_units()

    return cs ** 2 / (gamma * specific_gas_constant)


def velocity_radial_cylindrical(
    snap: Union[SnapLike, Sinks], origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the cylindrical radial velocity.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the cylindrical radial
        velocity as a Quantity like (x, y, z) * au. Default is
        (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial velocity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    x, y = pos[:, 0], pos[:, 1]
    vx, vy = vel[:, 0], vel[:, 1]

    return (x * vx + y * vy) / np.sqrt(x ** 2 + y ** 2)


def velocity_radial_spherical(
    snap: Union[SnapLike, Sinks], origin: Quantity = None, ignore_accreted: bool = False
) -> Quantity:
    """Calculate the spherical radial velocity.

    Parameters
    ----------
    snap
        The Snap object.
    origin : optional
        The origin around which to compute the spherical radial velocity
        as a Quantity like (x, y, z) * au. Default is (0, 0, 0).
    ignore_accreted : optional
        Ignore accreted particles. Default is False.

    Returns
    -------
    Quantity
        The radial velocity on the particles.
    """
    if ignore_accreted:
        h: Quantity = snap['smoothing_length']
        pos: Quantity = snap['position'][h > 0]
        vel: Quantity = snap['velocity'][h > 0]
    else:
        pos = snap['position']
        vel = snap['velocity']

    if origin is None:
        origin = ORIGIN
    pos = pos - origin

    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    vx, vy, vz = vel[:, 0], vel[:, 1], vel[:, 2]

    return (x * vx + y * vy + z * vz) / np.sqrt(x ** 2 + y ** 2 + z ** 2)



"""Add extra quantities to Snap."""

from .. import analysis

MODULES = {'common': analysis.particles, 'disc': analysis.discs}


def add_quantities(snap, category: str = 'common'):
    """Make extra quantities available on Snap.

    Parameters
    ----------
    snap
        The Snap object to add extra quantities to.
    category
        The category from which to add extra quantities. Options include:

        - 'common': adds common quantities appropriate for most
          simulations, such as angular momentum.
        - 'disc': adds accretion disc quantities, such as eccentricity
          or Keplerian frequency.

        Default is 'common'.
    """
    if category not in MODULES:
        raise ValueError(f'category must be in {list(MODULES)}')
    module = MODULES[category]

    for array in module.array_requires:  # type: ignore
        _add_array(snap, array, module)


def _can_add_array(snap, name, module):
    return set(module.array_requires[name]).issubset(snap.available_arrays())


def _vector(name, module):
    return True if name in module.vector_arrays else False


def _dust(name, module):
    return True if name in module.dust_arrays else False


def _add_array(snap, name, module):
    if _can_add_array(snap, name, module):
        vector = _vector(name, module)
        dust = _dust(name, module)
        snap._array_registry[name] = getattr(module, name)
        if vector is True:
            snap._vector_arrays.add(name)
        if dust is True:
            snap._dust_arrays.add(name)



"""Extra pre-defined profiles."""

from functools import partial

from .._logging import logger
from . import profiles


def extra_profiles(profile, num_separate_dust: int = 0, num_mixture_dust: int = 0):
    """Make extra profiles available.

    Parameters
    ----------
    profile
        The profile object to add extra profiles to.
    num_separate_dust
        The number of "separate sets of particles" dust species.
    num_mixture_dust
        The number of "mixture" dust species.
    """
    num_dust_species = num_mixture_dust + num_separate_dust

    for prof, ndims in profiles.appropriate_ndim.items():
        if profile.ndim in ndims:
            if prof not in profiles.dust_profiles:
                profile._profile_functions[prof] = getattr(profiles, prof)
            else:
                if profiles.dust_profiles[prof] == 'both':
                    for idx in range(num_dust_species):
                        profile._profile_functions[f'{prof}_{idx+1:03}'] = partial(
                            getattr(profiles, prof), idx
                        )
                elif profiles.dust_profiles[prof] == 'mixture':
                    for idx in range(num_mixture_dust):
                        profile._profile_functions[f'{prof}_{idx+1:03}'] = partial(
                            getattr(profiles, prof), idx
                        )
                elif profiles.dust_profiles[prof] == 'mixture (gas)':
                    if num_mixture_dust > 0:
                        profile._profile_functions[prof] = getattr(profiles, prof)
                else:
                    logger.debug(f'{prof} - cannot add profile')



"""Visualize SPH data.

The Plonk implementation for visualizing smoothed particle hydrodynamics
simulations using kernel density estimation based interpolation.
"""

from .animation import (
    animate,
    animation_images,
    animation_particles,
    animation_profiles,
)
from .interpolation import interpolate
from .simulation import visualize_sim
from .visualization import image, plot, vector

__all__ = [
    'animate',
    'animation_images',
    'animation_particles',
    'animation_profiles',
    'image',
    'interpolate',
    'plot',
    'vector',
    'visualize_sim',
]



"""Extra pre-defined profiles."""

import numpy as np

from .._units import Quantity
from .._units import units as plonk_units

G = (1 * plonk_units.newtonian_constant_of_gravitation).to_base_units()

# Some profiles are only appropriate in some geometries (Profile.ndim)
appropriate_ndim = {
    'alpha_shakura_sunyaev': [2],
    'angular_momentum_phi': [2, 3],
    'angular_momentum_theta': [2, 3],
    'aspect_ratio': [2],
    'disc_viscosity': [2],
    'dust_mass': [1, 2, 3],
    'dust_surface_density': [2],
    'epicyclic_frequency': [2],
    'gas_mass': [1, 2, 3],
    'gas_surface_density': [2],
    'linear_density': [1],
    'mass': [1, 2, 3],
    'midplane_stokes_number': [2],
    'scale_height': [2],
    'surface_density': [2],
    'toomre_q': [2],
}

# Dust profiles: can be 'mixture', 'mixture (gas)', 'separate', or 'both'
dust_profiles = {
    'dust_mass': 'mixture',
    'dust_surface_density': 'mixture',
    'gas_mass': 'mixture (gas)',
    'gas_surface_density': 'mixture (gas)',
    'midplane_stokes_number': 'both',
}


def alpha_shakura_sunyaev(prof) -> Quantity:
    """Shakura-Sunyaev alpha disc viscosity."""
    try:
        alpha = prof.snap._file_pointer['header/alpha'][()]
    except KeyError:
        raise ValueError('Cannot determine artificial viscosity alpha')
    return alpha / 10 * prof['smoothing_length'] / prof['scale_height']


def angular_momentum_phi(prof) -> Quantity:
    """Angle between specific angular momentum and x-axis in xy-plane."""
    angular_momentum_x = prof['angular_momentum_x']
    angular_momentum_y = prof['angular_momentum_y']
    return np.arctan2(angular_momentum_y, angular_momentum_x)


def angular_momentum_theta(prof) -> Quantity:
    """Angle between specific angular momentum and xy-plane."""
    angular_momentum_z = prof['angular_momentum_z']
    angular_momentum_magnitude = prof['angular_momentum_mag']
    return np.arccos(angular_momentum_z / angular_momentum_magnitude)


def aspect_ratio(prof) -> Quantity:
    """Aspect ratio profile."""
    H = prof['scale_height']
    R = prof['radius']
    return H / R


def disc_viscosity(prof) -> Quantity:
    """Disc viscosity. I.e. nu = alpha_SS c_s H."""
    return prof['alpha_shakura_sunyaev'] * prof['sound_speed'] * prof['scale_height']


def dust_mass(idx, prof) -> Quantity:
    """Dust mass profile."""
    M = prof.snap[f'dust_mass_{idx+1:03}']
    return prof.particles_to_binned_quantity('sum', M)


def dust_surface_density(idx, prof) -> Quantity:
    """Dust surface density profile.

    Units are [mass / length ** ndim], which depends on ndim of profile.
    """
    return prof[f'dust_mass_{idx+1:03}'] / prof['size']


def epicyclic_frequency(prof) -> Quantity:
    """Epicyclic frequency."""
    Omega = prof['keplerian_frequency']
    R = prof['radius']
    return np.sqrt(2 * Omega / R * np.gradient(R ** 2 * Omega, R))


def gas_mass(prof) -> Quantity:
    """Gas mass profile."""
    M = prof.snap['gas_mass']
    return prof.particles_to_binned_quantity('sum', M)


def gas_surface_density(prof) -> Quantity:
    """Gas surface density profile.

    Units are [mass / length ** ndim], which depends on ndim of profile.
    """
    return prof['gas_mass'] / prof['size']


def linear_density(prof) -> Quantity:
    """Linear density profile.

    Units are [mass / length].
    """
    if prof.ndim != 1:
        raise ValueError('linear_density is appropriate for linear profiles')
    return prof['mass'] / prof['size']


def mass(prof) -> Quantity:
    """Mass profile."""
    M = prof.snap['mass']
    return prof.particles_to_binned_quantity('sum', M)


def midplane_stokes_number(idx, prof) -> Quantity:
    """Midplane Stokes number profile."""
    gamma = prof.snap.properties['adiabatic_index']
    grain_density = prof.snap.properties['grain_density'][idx]
    grain_size = prof.snap.properties['grain_size'][idx]
    return (
        np.pi
        * np.sqrt(gamma)
        * grain_density
        * grain_size
        / 2
        / prof['surface_density']
    )


def scale_height(prof) -> Quantity:
    """Scale height profile."""
    z = prof.snap['z']
    return prof.particles_to_binned_quantity('std', z)


def surface_density(prof) -> Quantity:
    """Surface density profile.

    Units are [mass / length ** 2].
    """
    if prof.ndim != 2:
        raise ValueError('line_density is appropriate for linear profiles')
    return prof['mass'] / prof['size']


def toomre_q(prof) -> Quantity:
    """Toomre Q parameter."""
    return (
        prof['sound_speed']
        * prof['keplerian_frequency']
        / (np.pi * G * prof['surface_density'])
    )



"""Calculate global (total) quantities on the particles."""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

import numpy as np
from numpy import ndarray

from .._units import Quantity
from ..utils.math import norm
from . import particles

if TYPE_CHECKING:
    from ..snap.snap import SnapLike, SubSnap


def accreted_mass(snap: SnapLike) -> Quantity:
    """Calculate the accreted mass.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    Quantity
        The accreted mass.
    """
    h: Quantity = snap['smoothing_length']
    _mass: Quantity = snap['mass'][~(h > 0)]

    return _mass.sum()


def angular_momentum(
    snap: SnapLike, sinks: Union[bool, List[int]] = True, origin: Quantity = None
) -> Quantity:
    """Calculate the total angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).
    origin : optional
        The origin around which to compute the angular momentum as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).

    Returns
    -------
    Quantity
        The total angular momentum like (lx, ly, lz).
    """
    _sinks = _get_sinks(snap, sinks)

    angmom = particles.angular_momentum(
        snap=snap, origin=origin, ignore_accreted=True
    ).sum(axis=0)

    if len(_sinks) > 0:
        angmom_sinks = particles.angular_momentum(
            snap=_sinks, origin=origin, ignore_accreted=False
        )
        if angmom_sinks.ndim == 1:
            angmom = angmom + angmom_sinks
        elif angmom_sinks.ndim == 2:
            angmom = angmom + angmom_sinks.sum(axis=0)

    return angmom


def center_of_mass(snap: SnapLike, sinks: Union[bool, List[int]] = True) -> Quantity:
    """Calculate the center of mass.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    Quantity
        The center of mass as a vector (cx, cy, cz).
    """
    _sinks = _get_sinks(snap, sinks)

    h: Quantity = snap['smoothing_length']
    _mass: Quantity = snap['mass'][h > 0]
    pos: Quantity = snap['position'][h > 0]

    mass_pos = (_mass[:, np.newaxis] * pos).sum(axis=0)
    mass_tot = _mass.sum()

    if len(_sinks) > 0:
        mass_sinks = _sinks['mass']
        pos_sinks = _sinks['position']
        if pos_sinks.ndim == 1:
            mass_pos = mass_pos + mass_sinks * pos_sinks
            mass_tot = mass_tot + mass_sinks
        elif pos_sinks.ndim == 2:
            mass_pos = mass_pos + (mass_sinks[:, np.newaxis] * pos_sinks).sum(axis=0)
            mass_tot = mass_tot + mass_sinks.sum()

    return mass_pos / mass_tot


def dust_mass(snap: SnapLike, squeeze: Union[bool, List[int]] = False) -> Quantity:
    """Calculate the total dust mass per species.

    Parameters
    ----------
    snap
        The Snap object.
    squeeze
        If True return all subtypes in a single array. Default is
        False.

    Returns
    -------
    Quantity
        The total dust mass per species.
    """
    if snap.num_dust_species == 0:
        raise ValueError('No dust available')
    elif snap.num_dust_species > 0:
        if 'dust_fraction' in snap.available_arrays():
            h: Quantity = snap['smoothing_length']
            _mass: Quantity = snap['mass'][h > 0]
            dustfrac: Quantity = snap['dust_fraction'][h > 0]
            dustmass = (_mass[:, np.newaxis] * dustfrac).sum(axis=0)
            if squeeze:
                return dustmass.sum()
            return dustmass
        if snap.num_particles_of_type.get('dust') is not None:
            subsnaps: List[SnapLike] = snap.family('dust')  # type: ignore
            dustmass = [
                subsnap['mass'][subsnap['smoothing_length'] > 0].sum()
                for subsnap in subsnaps
            ]
            if squeeze:
                _dustmass = dustmass[0]
                for val in dustmass[1:]:
                    _dustmass = _dustmass + val
                return _dustmass
            return dustmass
    raise ValueError('Cannot calculate dust mass')


def gas_mass(snap: SnapLike) -> Quantity:
    """Calculate the total gas mass.

    Parameters
    ----------
    snap
        The Snap object.

    Returns
    -------
    Quantity
        The total gas mass.
    """
    gas: SubSnap = snap.family('gas')  # type: ignore
    if snap.num_dust_species == 0:
        return mass(gas, sinks=False)
    elif snap.num_dust_species > 0:
        if 'dust_fraction' in snap.available_arrays():
            h: Quantity = snap['smoothing_length']
            _mass: Quantity = snap['mass'][h > 0]
            dustfrac: Quantity = snap['dust_fraction'][h > 0]
            dustmass = (_mass[:, np.newaxis] * dustfrac).sum()
            mixturemass = mass(gas, sinks=False)
            return mixturemass - dustmass
        else:
            return mass(gas, sinks=False)
    raise ValueError('Cannot calculate gas mass')


def kinetic_energy(snap: SnapLike, sinks: Union[bool, List[int]] = True) -> Quantity:
    """Calculate the total kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    Quantity
        The total kinetic energy.
    """
    _sinks = _get_sinks(snap, sinks)

    ke = particles.kinetic_energy(snap=snap, ignore_accreted=True).sum()

    if len(_sinks) == 1:
        m, v = _sinks['mass'], _sinks['velocity']
        ke = ke + 1 / 2 * m * norm(v) ** 2
    elif len(_sinks) > 1:
        ke = ke + particles.kinetic_energy(snap=_sinks).sum()

    return ke


def mass(snap: SnapLike, sinks: Union[bool, List[int]] = True) -> Quantity:
    """Calculate the total mass.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    Quantity
        The total mass.
    """
    _sinks = _get_sinks(snap, sinks)

    h: Quantity = snap['smoothing_length']
    mass: Quantity = snap['mass'][h > 0]

    mass_sum = mass.sum()
    if len(_sinks) > 0:
        mass_sum = mass_sum + np.sum(_sinks['mass'])

    return mass_sum


def momentum(snap: SnapLike, sinks: Union[bool, List[int]] = True) -> Quantity:
    """Calculate the total momentum.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    Quantity
        The total linear momentum like (px, py, pz).
    """
    _sinks = _get_sinks(snap, sinks)

    mom = particles.momentum(snap=snap, ignore_accreted=True).sum(axis=0)

    if len(_sinks) == 1:
        mom = mom + _sinks['mass'] * _sinks['velocity']
    elif len(_sinks) > 1:
        mass = _sinks['mass']
        velocity = _sinks['velocity']
        mom = mom + (mass[:, np.newaxis] * velocity).sum(axis=0)

    return mom


def specific_angular_momentum(
    snap: SnapLike, sinks: Union[bool, List[int]] = True, origin: Quantity = None
) -> Quantity:
    """Calculate the total specific angular momentum.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).
    origin : optional
        The origin around which to compute the angular momentum as a
        Quantity like (x, y, z) * au. Default is (0, 0, 0).

    Returns
    -------
    Quantity
        The total specific angular momentum on the particles like
        (hx, hy, hz).
    """
    _sinks = _get_sinks(snap, sinks)

    angmom = particles.specific_angular_momentum(
        snap=snap, origin=origin, ignore_accreted=True
    ).sum(axis=0)

    if len(_sinks) > 0:
        angmom_sinks = particles.specific_angular_momentum(
            snap=_sinks, origin=origin, ignore_accreted=False
        )
        if angmom_sinks.ndim == 1:
            angmom = angmom + angmom_sinks
        elif angmom_sinks.ndim == 2:
            angmom = angmom + angmom_sinks.sum(axis=0)

    return angmom


def specific_kinetic_energy(
    snap: SnapLike, sinks: Union[bool, List[int]] = True
) -> Quantity:
    """Calculate the total specific kinetic energy.

    Parameters
    ----------
    snap
        The Snap object.
    sinks : optional
        Include sink particles specified by a list of indices, or a
        bool indicating all sinks or no sinks. Default is True (all
        sinks).

    Returns
    -------
    Quantity
        The total specific kinetic energy.
    """
    _sinks = _get_sinks(snap, sinks)

    ke = particles.specific_kinetic_energy(snap=snap, ignore_accreted=True).sum()

    if len(_sinks) == 1:
        v = _sinks['velocity']
        ke = ke + 1 / 2 * norm(v) ** 2
    elif len(_sinks) > 1:
        ke = ke + particles.specific_kinetic_energy(snap=_sinks).sum()

    return ke


def _get_sinks(snap, sinks):
    if sinks is False:
        return []
    if snap.num_sinks > 0:
        if sinks is True:
            return snap.sinks
        elif isinstance(sinks, (int, tuple, list, ndarray)):
            return snap.sinks[sinks]
    return []



"""Reader for snapshot files."""

# To write a new reader, three functions are required:
#
# - snap_properties_and_units,
# - snap_array_registry,
# - snap_sink_registry.
#
# These functions take an h5py.File object as an argument (and optionally
# a name map dictionary to rename arrays on file) and a string data_source
# to say where the data is from, e.g. 'phantom'.
#
# The first function gets some properties and code units, e.g. from the
# Phantom header. It is called by Snap.load_snap as follows
#
# self._properties, self._code_units = snap_properties_and_units(
#     file_pointer=self._file_pointer, data_source=self.data_source
# )
#
# The second two functions each create a dictionary with
# keys as names of arrays and values as a function that when called with a
# Snap object returns that array. They are called by Snap.load_snap as follows
#
# self._array_registry.update(
#     snap_array_registry(
#         file_pointer=self._file_pointer,
#         data_source=self.data_source,
#         name_map=self._name_map['particles'],
#     )
# )
#
# self._sink_registry.update(
#     snap_sink_registry(
#         file_pointer=self._file_pointer,
#         data_source=self.data_source,
#         name_map=self._name_map['sinks'],
#     )
# )
#
# See phantom.py for an example.

from __future__ import annotations

from typing import Any, Callable, Dict, Tuple

import h5py

from ..._units import Quantity
from .phantom import snap_array_registry as snap_array_registry_phantom
from .phantom import snap_properties_and_units as snap_properties_and_units_phantom
from .phantom import snap_sink_registry as snap_sink_registry_phantom

DATA_SOURCES = ['phantom']


def snap_properties_and_units(
    file_pointer: h5py.File, data_source: str
) -> Tuple[Dict[str, Any], Dict[str, Quantity]]:
    """Generate snap properties and units.

    Parameters
    ----------
    file_pointer
        The h5py file pointer to the snap file.
    data_source
        The SPH code that produced the data. E.g. 'phantom'.

    Returns
    -------
    properties
        The properties as a dict.
    code_units
        The code units as a dict.
    """
    # The properties dictionary has keys like "equation_of_state" and "time", with
    # varied values like "locally isothermal disc" and "61485663602.558136
    # <Unit('second')>".

    # The code_units dictionary has keys like "length" and "mass" with values as Pint
    # quantities, e.g. "149600000000.0 <Unit('meter')>" and "1.9891000000000002e+30
    # <Unit('kilogram')>".

    if data_source.lower() == 'phantom':
        return snap_properties_and_units_phantom(file_pointer=file_pointer)
    raise RuntimeError('Cannot generate properties and units')


def snap_array_registry(
    file_pointer: h5py.File,
    data_source: str,
    name_map: Dict[str, str] = None,
) -> Dict[str, Callable]:
    """Generate snap array registry.

    Parameters
    ----------
    file_pointer
        The h5py file pointer to the snap file.
    data_source
        The SPH code that produced the data. E.g. 'phantom'.
    name_map : optional
        A dict to convert from Phantom array names to Plonk names.

    Returns
    -------
    Dict
        The particle array registry.
    """
    # The keys are the names of the particle arrays and the values are functions that
    # return the array when called with snap as the argument.

    if data_source.lower() == 'phantom':
        return snap_array_registry_phantom(file_pointer=file_pointer, name_map=name_map)
    raise RuntimeError('Cannot generate array registry')


def snap_sink_registry(
    file_pointer: h5py.File,
    data_source: str,
    name_map: Dict[str, str] = None,
) -> Dict[str, Callable]:
    """Generate snap sink registry.

    Parameters
    ----------
    file_pointer
        The h5py file pointer to the snap file.
    data_source
        The SPH code that produced the data. E.g. 'phantom'.
    name_map : optional
        A dict to convert from Phantom array names to Plonk names.

    Returns
    -------
    Dict
        The sink array registry.
    """
    # The keys are the names of the sink arrays and the values are functions that return
    # the array when called with snap as the argument.

    if data_source.lower() == 'phantom':
        return snap_sink_registry_phantom(file_pointer=file_pointer, name_map=name_map)
    raise RuntimeError('Cannot generate sink registry')
